/* pptp_server.c ... Call manager for PPTP connections.
 *                    Handles TCP port 1723 protocol.
 *                    C. Scott Ananian <cananian@alumni.princeton.edu>
 *
 * $Id: pptp_server.c,v 1.8 1997/12/15 12:25:00 cananian Exp $
 */
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <setjmp.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include "pptp_callmgr.h"
#include "pptp_ctrl.h"
#include "pptp_msg.h"
#include "dirutil.h"
#include "vector.h"
#include "util.h"

#define SESSION_ALARM  (180*60);  /* (session length in seconds) */
 
int open_inetsock(struct in_addr *inetaddr);
int open_unixsock(struct in_addr inetaddr);
void close_inetsock(int fd, struct in_addr inetaddr);
void close_unixsock(int fd, struct in_addr inetaddr);

sigjmp_buf env;
void sighandler(int sig) {
  warn("received signal %d", sig);
  siglongjmp(env, 1);
}

struct local_callinfo {
  int unix_sock;
  pid_t pid[2];
};

struct local_conninfo {
  VECTOR * call_list;
  fd_set * call_set;
  int *max_fd;
  int *unix_sock;
};

void do_nothing(int sig) { /* do nothing signal handler */ }

int MainArgc;
struct in_addr inetaddr;
char **MainArgv;

/* Connection callback */
void conn_callback(PPTP_CONN *conn, enum conn_state state) {

  switch(state) {
  case CONN_OPEN_FAIL:
  case CONN_CLOSE_DONE:
    /* get outta here */
    siglongjmp(env, 1);
    break;
  default:
    log("Unhandled connection callback state [%d].", (int) state);
    break;
  }
}
/* Call callback */
void call_callback(PPTP_CONN *conn, PPTP_CALL *call, enum call_state state) {
  struct local_callinfo *lci;
  struct local_conninfo *conninfo;
  u_int16_t call_id[2];

  switch(state) {
  case CALL_OPEN_DONE:
    /* okey dokey.  This means that the call_id and peer_call_id are now
     * valid, so lets send them on to our friends who requested this call.
     */
    lci = pptp_call_closure_get(conn, call);
    conninfo = pptp_conn_closure_get(conn);
     
    /* This means the server is creating a new connection. We fork off
       pptp proper and tell it to 'connect' to the address that connected to
       us. It will see our socket in /var/run/pptp and connect to it. We
       can then tell it the connection is established. */
    if (lci == NULL)
    {	
      struct sockaddr_un from;
      int len = sizeof(from);
      int s;
      int retval;

      /* Step 1b: FORK and return status to calling process. */
      switch (fork()) 
      {
         case 0: /* child. stick around. */
	 {
	   char *Args[300];
	   int I;
	   for (I = 2; I - 1 < MainArgc && I < 300; I++)
	       Args[I] = MainArgv[I-1];
		
	   Args[I] = 0;
	   Args[0] = PPTP_PPTP_BINARY;
	   Args[1] = inet_ntoa(inetaddr);
	    
	   // Close all fds and redirect stderr, stdout, stdin to /dev/null
	   for (I = 0; I != 10; I++)
	       close(I);
	   open("/dev/null",O_RDWR);
	   open("/dev/null",O_RDWR);
	   open("/dev/null",O_RDWR);
	    
	   execv(PPTP_PPTP_BINARY,Args);
	   fatal("Coult not exec");
	  break;
	 }
	 
	 case -1: /* failure.  Fatal. */
	  fatal("Could not fork.");
	 default: /* Parent. */
	 break;
      }       
       
      /* Accept the socket */
      if ((s = accept(*conninfo->unix_sock, (struct sockaddr *) &from, &len))<0) {
	warn("Socket not accepted: %s", strerror(errno));
        exit(0);
      }
      /* Allocate memory for local call information structure. */
      if ((lci = malloc(sizeof(*lci))) == NULL) {
	warn("Out of memory."); close(s); exit(0);
      }
      lci->unix_sock = s;

      /* Read and store the associated pids */
      read(s, &lci->pid[0], sizeof(lci->pid[0]));
      read(s, &lci->pid[1], sizeof(lci->pid[1]));
      /* associate the local information with the call */
      pptp_call_closure_put(conn, call, (void *) lci);

       /* Keep alive; wait for close */
      retval = vector_insert(conninfo->call_list, s, call); assert(retval);
      if (s > *conninfo->max_fd) *conninfo->max_fd = s;
      FD_SET(s, conninfo->call_set);
    }
     
    pptp_call_get_ids(conn, call, &call_id[0], &call_id[1]);
    write(lci->unix_sock, &call_id, sizeof(call_id));
    /* Our duty to the fatherland is now complete. */
    break;
  case CALL_OPEN_FAIL:
  case CALL_CLOSE_RQST:
  case CALL_CLOSE_DONE:
    /* don't need to do anything here, except make sure tables are sync'ed */
    conninfo = pptp_conn_closure_get(conn);
    lci = pptp_call_closure_get(conn, call);
    if (lci == 0 || conninfo == 0)
	break;
    if (vector_contains(conninfo->call_list, lci->unix_sock)) {
      vector_remove(conninfo->call_list, lci->unix_sock);
      close(lci->unix_sock);
      FD_CLR(lci->unix_sock, conninfo->call_set);
      kill(lci->pid[0], SIGTERM);
      kill(lci->pid[1], SIGTERM);
    }
    break;
  default:
    log("Unhandled call callback state [%d].", (int) state);
    break;
  }
}

/* Call Manager */
int main(int argc, char **argv, char **envp) {
  int inet_sock, unix_sock;
  fd_set call_set;
  PPTP_CONN * conn;
  VECTOR * call_list;
  int max_fd;
  int first;
  int retval;
  int i;
  time_t timer;
   
  MainArgc = argc;
  MainArgv = argv;

  /* Step 1: Open sockets. */
  if ((inet_sock = open_inetsock(&inetaddr)) < 0)
    fatal("Could not open control connection to %s", argv[1]);
  if ((unix_sock = open_unixsock(inetaddr)) < 0)
    fatal("Could not open unix socket for %s", argv[1]);

  /* Step 1c: Clean up unix socket on TERM */
  signal(SIGINT, sighandler);
  signal(SIGTERM, sighandler);
  signal(SIGKILL, sighandler);
  if (sigsetjmp(env, 1)!=0) goto cleanup;
  signal(SIGPIPE, do_nothing);
  signal(SIGUSR1, do_nothing); /* signal state change; wake up accept */

  /* Step 2: Open control connection and register callback */
  if ((conn = pptp_conn_open(inet_sock, 0, NULL/* callback */)) == NULL) {
    close(unix_sock); close(inet_sock); fatal("Could not open connection.");
  }
  pptp_server_callback(conn,call_callback);

  FD_ZERO(&call_set);
  max_fd = unix_sock;
  if (inet_sock > unix_sock)
      max_fd = inet_sock;
  call_list = vector_create();
  { 
    struct local_conninfo *conninfo = malloc(sizeof(*conninfo));
    if (conninfo==NULL) {
      close(unix_sock); close(inet_sock); fatal("No memory.");
    }
    conninfo->call_list = call_list;
    conninfo->call_set  = &call_set;
    conninfo->max_fd = &max_fd;
    conninfo->unix_sock = &unix_sock;
    pptp_conn_closure_put(conn, conninfo);
  }

  if (sigsetjmp(env, 1)!=0) goto shutdown;
  timer = time(0) + SESSION_ALARM;

  /* Step 3: Get FD_SETs */
  first = 1;
  do {
    fd_set read_set, write_set, excpt_set = call_set;
    FD_ZERO(&read_set);
    FD_ZERO(&write_set);
    FD_SET(unix_sock, &read_set);
    pptp_fd_set(conn, &read_set, &write_set);

    /* Step 4: Wait on INET or UNIX event */
    if (select(max_fd+1, &read_set, &write_set, &excpt_set, NULL)<0)
      /* a signal or somesuch. */
      continue;

    /* Step 5a: Handle INET events */
    pptp_dispatch(conn, &read_set, &write_set);
    if (pptp_reap(conn) != 0)
	goto shutdown;
     
    /* Step 5b: Handle new connection to UNIX socket */
    if (FD_ISSET(unix_sock, &read_set)) {
      /* New call! */
      struct sockaddr_un from;
      int len = sizeof(from);
      PPTP_CALL * call;
      struct local_callinfo *lci;
      int s;

      /* Accept the socket */
      if ((s = accept(unix_sock, (struct sockaddr *) &from, &len))<0) {
	warn("Socket not accepted: %s", strerror(errno));
	goto skip_accept;
      }
      /* Allocate memory for local call information structure. */
      if ((lci = malloc(sizeof(*lci))) == NULL) {
	warn("Out of memory."); close(s); goto skip_accept;
      }
      lci->unix_sock = s;

      /* Give the initiator time to write the PIDs while we open the call */
      call = pptp_call_open(conn, call_callback);
      /* Read and store the associated pids */
      read(s, &lci->pid[0], sizeof(lci->pid[0]));
      read(s, &lci->pid[1], sizeof(lci->pid[1]));
      /* associate the local information with the call */
      pptp_call_closure_put(conn, call, (void *) lci);
      /* The rest is done on callback. */
      
      /* Keep alive; wait for close */
      retval = vector_insert(call_list, s, call); assert(retval);
      if (s > max_fd) max_fd = s;
      FD_SET(s, &call_set);
      first = 0;
    }
  skip_accept:
    /* Step 5c: Handle socket close */
    for (i=0; i<max_fd; i++)
      if (FD_ISSET(i, &excpt_set)) {
	/* close it */
	PPTP_CALL * call;
	retval = vector_search(call_list, i, &call);
	if (retval) {
	  struct local_callinfo *lci = pptp_call_closure_get(conn, call);
	  kill(lci->pid[0], SIGTERM);
	  kill(lci->pid[1], SIGTERM);
	  free(lci);
	  /* soft shutdown.  Callback will do hard shutdown later */
	  pptp_call_close(conn, call);
	  vector_remove(call_list, i);
	}
	FD_CLR(i, &call_set);
	close(i);
      }
  } while ( (vector_size(call_list)>0 || first) && (time(0) <= timer) );
  
  if (time(0) > timer)
    warn("session expired");

shutdown:
  {
    fd_set read_set, write_set;

    /* kill all open calls */
    for (i=0; i<vector_size(call_list); i++) {
      PPTP_CALL *call = vector_get_Nth(call_list, i);
      struct local_callinfo *lci = pptp_call_closure_get(conn, call);
      pptp_call_close(conn, call);
      kill(lci->pid[0], SIGTERM);
      kill(lci->pid[1], SIGTERM);
    }
    /* attempt to dispatch these messages */
    FD_ZERO(&read_set);
    FD_ZERO(&write_set);
    pptp_fd_set(conn, &read_set, &write_set);
    FD_ZERO(&read_set);
    pptp_dispatch(conn, &read_set, &write_set);
    if (i>0) sleep(2);
    /* no more open calls.  Close the connection. */
    pptp_conn_close(conn, PPTP_STOP_LOCAL_SHUTDOWN);
    FD_ZERO(&read_set);
    FD_ZERO(&write_set);
    pptp_fd_set(conn, &read_set, &write_set);
    FD_ZERO(&read_set);
    pptp_dispatch(conn, &read_set, &write_set);
    sleep(2);
    /* with extreme prejudice */
    pptp_conn_destroy(conn);
    vector_destroy(call_list);
  }
cleanup:
  close_inetsock(inet_sock, inetaddr);
  close_unixsock(unix_sock, inetaddr);
  return 0;
}

int open_inetsock(struct in_addr *inetaddr) {
  struct sockaddr_in Name;
  unsigned int Sz;

#ifdef SERVERMODE
   int s;
  int Conn;
  int True = 1;

  memset(&Name,0,sizeof(Name));
  Name.sin_family = AF_INET;
  Name.sin_port   = htons(PPTP_PORT);

  if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    warn("socket: %s", strerror(errno));
    close(s);
    return -1;
  }
  setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(char *)&True,sizeof(True));
  if (bind(s,(struct sockaddr *)&Name,sizeof(Name)) != 0) {
    warn("bind: %s", strerror(errno));
    close(s);
    return -1;
  }   
      
  if (listen(s,1) != 0) {
    warn("listen: %s", strerror(errno));
    close(s); 
    return -1;
  }
   
  Sz = sizeof(Name);
  Conn = accept(s,(struct sockaddr *)&Name,&Sz);
  close(s);
  memcpy(inetaddr,&Name.sin_addr,sizeof(*inetaddr));

  return Conn;
#else
  Sz = sizeof(Name);
  getpeername(0,(struct sockaddr *)&Name,&Sz);
  memcpy(inetaddr,&Name.sin_addr,sizeof(*inetaddr));

  return 0;
#endif   
}
int open_unixsock(struct in_addr inetaddr) {
  struct sockaddr_un where;
  struct stat st;
  char *dir;
  int s;

  if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
    warn("socket: %s", strerror(errno));
    return s;
  }

  where.sun_family = AF_UNIX;
  snprintf(where.sun_path, sizeof(where.sun_path), 
	   PPTP_SOCKET_PREFIX "%s", inet_ntoa(inetaddr));

  if (stat(where.sun_path, &st) >= 0) {
    warn("Call manager for %s is already running.", inet_ntoa(inetaddr));
    close(s); return -1;
  }

  /* Make sure path is valid. */
  dir = dirname(where.sun_path);
  if (!make_valid_path(dir, 0770))
    fatal("Could not make path to %s: %s", where.sun_path, strerror(errno));
  free(dir);

  if (bind(s, (struct sockaddr *) &where, sizeof(where)) < 0) {
    warn("bind: %s", strerror(errno));
    close(s); return -1;
  }

  chmod(where.sun_path, 0777);

  listen(s, 127);

  return s;
}
void close_inetsock(int fd, struct in_addr inetaddr) {
  close(fd);
}
void close_unixsock(int fd, struct in_addr inetaddr) {
  struct sockaddr_un where;
  close(fd);
  snprintf(where.sun_path, sizeof(where.sun_path), 
	   PPTP_SOCKET_PREFIX "%s", inet_ntoa(inetaddr));
  unlink(where.sun_path);
}
  




