/**
 * \file date.c
 * \brief time and calendar functions
 * \version $Revision:$
 * \date Last change: $Date:$
 * \author Oliver Dawid
 * \author Last change by: $Author:$
 */

/*----------------------------------------------------------------------------
 *  circd server for Linux
 *
 *  Copyright (c) 2005-2014 - Oliver Dawid, fli4l-Team <team@fli4l.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *----------------------------------------------------------------------------
 */

#include "circd.h"

#define N_HOLIDAYS 9

void add_days(int start_day, int start_month, int start_year, int n_days, int *new_day_p, int *new_month_p);

/** describes one day of a year */
struct holidays
{
	int		day;	/**< The day of the month (1 == 1.) */
	int		month;	/**< The month (1 == January) */
};

/** dynamic struct for holidays of current year */
struct holidays holidays[N_HOLIDAYS];

/** This is a static struct for global definition of holidays. If the month is
* == 0, this denotes an event \a day days after easter.  */
struct holidays holiday_table[N_HOLIDAYS] = {
	{  1,  1 },				    /* Neujahr			    */
	{  1,  5 },				    /* Maifeiertag		    */
	{  3, 10 },				    /* Tag der deutschen Einheit    */
	{ 25, 12 },				    /* 1. Weihnachtsfeiertag	    */
	{ 26, 12 },				    /* 2. Weihnachtsfeiertag	    */
	{ -2,  0 },				    /* Karfreitag		    */
	{  1,  0 },				    /* Ostermontag		    */
	{ 39,  0 },				    /* Christi Himmelfahrt	    */
	{ 50,  0 },				    /* Pfingstmontag		    */
#if 0
	{ 60,  0 },				    /* Fronleichnam		    */
#endif
};

/** enumeration of the days per month */
static int  g_days_per_month[]    =
{
	31,					/* January			    */
	28,					/* February (may be changed to 29)  */
	31,					/* March			    */
	30,					/* April			    */
	31,					/* May				    */
	30,					/* June				    */
	31,					/* July				    */
	31,					/* August			    */
	30,					/* September			    */
	31,					/* October			    */
	30,					/* November			    */
	31					/* December			    */
};

/** calulate if \a y is a leap year */
#define IS_LEAP_YEAR(y)	((((y) % 4) == 0) && ((((y) % 100) != 0) || (((y) % 400) == 0)))

/**
 * This function checks if the supplied day is a holiday.
 * 
 * \todo make the local vars more self-explain
 *
 * \param day the day to check (1 == 1.)
 * \param month the month to check (1 == January)
 * \param year the year to check (4-digit)
 *
 * \retval TRUE day is a holiday
 * \retval FALSE day is not a holiday
 */
int 
is_holiday(int day, int month, int year)
{
	static int last_year;
	static int ostern_day, ostern_month;
	int a, b, c, d, e, m, s, M, N, D, i, offset;
	
	if (last_year != year) {
		last_year = year;
		
		a = year % 19;
		b = year % 4;
		c = year % 7;
		
		m = ((8 * (year/100) + 13) / 25) - 2;
		s = (year/100) - (year/400) - 2;
		
		M = (15 + s - m) % 30;
		N = (6 + s) % 7;
		
		d = (M + 19 * a) % 30;
		
		if (d == 29) {
			D = 28;
		} else if (d == 28 && a >= 11) {
			D = 27;
		} else {
			D = d;
		}
		
		e = (2 * b + 4 * c + 6 * D + N) % 7;
		
		offset = D + e + 1;
		
		ostern_day = 21 + offset;
		ostern_month = 3;
		
		while (ostern_day > 31) {
			ostern_day -= 31;
			ostern_month++;
		}
		
		for (i = 0; i < N_HOLIDAYS; i++) {
			if (holiday_table[i].month == 0) {
				add_days(ostern_day, ostern_month, year, holiday_table[i].day, &(holidays[i].day), &(holidays[i].month));
			} else {
				holidays[i].day	    = holiday_table[i].day;
				holidays[i].month   = holiday_table[i].month;
			}
		}
#ifdef TEST
		printf ("Ostern: %02d.%02d.%04d\n", ostern_day, ostern_month, year);
#endif
	}
	
	for (i = 0; i < N_HOLIDAYS; i++) {
		if (holidays[i].day == day && holidays[i].month == month) {
			return TRUE;
		}
	}
	return FALSE;
}

/**
* This function adds a specific number of days to a given date.
 *
 * \warning Since the year is not returned, it would not be a wise choice to
 * set \a n_days to more then 365 or less then -365, since there is no way to
 * find out in which year you landed...
 *
 * \param start_day day of month of the starting date (1 == 1.)
 * \param start_month month of year of the starting date (1 == January)
 * \param start_year year of the starting date (4-digit)
 * \param n_days number of days to add (or substract if negative)
 * \param new_day_p pointer to new value of day
 * \param new_month_p pointer to new value of month
 */
void 
add_days(int start_day, int start_month, int start_year, int n_days, int *new_day_p, int *new_month_p)
{
	int days_this_month;
	
	while (n_days > 0) {
		start_day += n_days;
		
		days_this_month = g_days_per_month[start_month - 1];
		
		if (start_month == 1 && IS_LEAP_YEAR(start_year)) {
			days_this_month++;
		}
		
		if (start_day > days_this_month) {
			n_days	= start_day - days_this_month - 1;
			start_day = 1;
			start_month++;
			
			if (start_month > 12) {
				start_month = 1;
				start_year++;
			}
		} else {
			n_days = 0;
		}
	}
	
	while (n_days < 0) {
		start_day += n_days;
		
		if (start_day <= 0) {
			n_days = start_day;
			
			start_month--;
			
			if (start_month == 0) {
				start_month = 12;
				start_year--;
			}
			
			days_this_month = g_days_per_month[start_month - 1];
			
			if (start_month == 1 && IS_LEAP_YEAR(start_year)) {
				days_this_month++;
			}
			
			start_day = days_this_month;
		} else {
			n_days = 0;
		}
	}
	
	*new_day_p = start_day;
	*new_month_p = start_month;
	
	return;
}
