/*!
 * $Id: hwsuppd.c 34658 2014-11-06 23:10:34Z kristov $
 *
 * @file        hwsuppd.h
 * @brief       implementation of hwsuppd deamon
 * @details     program to control push button switches
 *
 *              see @ref usage for command line parameters
 *
 * @author  Carsten Spie&szlig; fli4l@carsten-spiess.de
 *          and the fli4l developer-team
 * @date    17.08.2013
 *
 * @note
 * portions based on
 * (c) 2005 by Torsten Georg <fli4l<at>foddy.de>
 *
 * (c) Copyright 2006 Stefan Krister <stefan@fli4l.de>
 * and the fli4l developer-team, All Rights Reserved.
 *
 * @note
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * Version 2 as published by the Free Software Foundation.
 *
 * I don't accept liability nor provide warranty for this Software.
 * This material is provided "AS-IS" and at no charge.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>

#include "hwsuppd.h"

static void usage (char *argv);
static int parse_args (int argc, char **argv);
static void main_loop (void);
static void sig_handler(int signr);
static void check_for_zombies (void);

/*!
 * @var         g_data
 * @brief       global data structure
 */
dataT g_data = {
        .nButton = -1,
        .fExecute = 1,
        .fForeground = 0,
        .fDaemonized = 0,
        .fDebug = 0
};



/*!
 * @fn          static int usage (char *argv)
 * @brief       print usage information
 * @param[in]   argv
 *              programm name
 */
static void usage (char *argv)
{
    printf("--------------------------------------------------------------\n");
    printf(" hwsuppd usage\n\n");
    printf(" Usage: %s [-b <n>] [-d] [-f]\n",
            argv);
    printf("  -b <n>: number of buttons \n");
    printf("  -d:     debug\n");
    printf("  -f:     foreground\n");
    printf("--------------------------------------------------------------\n");
}

/*!
 * @fn          int main (int argc, char **argv)
 * @brief       programm entry function
 * @param[in]   argc
 *              number of commandline arguments
 * @param[in]   argv
 *              commandline arguments
 * @return      result code
 */
int main (int argc, char **argv)
{
    int result;
    int i;

    result =  parse_args (argc, argv);
    if (result < 0)
    {
        goto exit;
    }

    if (!g_data.fForeground)
    {
        if (daemon(1, 1) == -1)
        {
            log_printf(LOG_ERR, "cannot detach");
            exit(1);
        }
        g_data.fDaemonized = 1;
    }

    openlog("hwsuppd", LOG_PID, LOG_USER);

    for (i = 0; i < g_data.nButton; i++)
    {
        g_data.statusButton[i] = 1;
    }

    if (0 == result)
    {
        main_loop ();
    }

exit:
    return result;
}

/*!
 * @fn          static void main_loop ()
 * @brief       main programm loop
 * @details     read buttons and execute external command
 */
static void main_loop (void)
{
    int i;

    log_printf (LOG_INFO, "main loop entered");

    signal (SIGINT, sig_handler);
    signal (SIGTERM, sig_handler);

    while (g_data.fExecute)
    {
        check_for_zombies();

        // handle buttons
        for (i = 0; i < g_data.nButton; i++)
        {
            button_read (i);
        }
        usleep(10000); // wait 10ms
    }
    log_printf (LOG_INFO, "main loop finished");
}

/*!
 * @fn          static void sig_handler(int signr)
 * @brief       signal handler
 * @details     terminate main loop on SIGINT and SIGTERM
 * @param       signr
 *              signal number
 */
static void sig_handler(int signr)
{
    log_printf (LOG_INFO, "signal %d received",signr);
    switch (signr)
    {
        case SIGINT:
        case SIGTERM:
            g_data.fExecute = 0;
            break;
    }
}

/*!
 * @fn          static int parse_args (int argc, char **argv)
 * @brief       parse command line arguments
 * @param[in]   argc
 *              number of commandline arguments
 * @param[in]   argv
 *              commandline arguments
 * @return      result code
 * @retval      0
 *              o.k.
 * @retval      -1
 *              failure, exit program
 */
static int parse_args (int argc, char **argv)
{
    int result = 0;
    char c;

    while ((c = getopt(argc, argv, "fdib:")) != -1)
    {
        switch (c)
        {
            case 'b':
                g_data.nButton = atoi(optarg);
                if (g_data.nButton > MAX_BUTTON)
                {
                    g_data.nButton = MAX_BUTTON;
                }
                break;
            case 'd':
                g_data.fDebug = 1;
                break;
            case 'f':
                g_data.fForeground = 1;
                break;
            default:
                usage (argv[0]);
                result = -1;
                break;
        }
    }


    if (optind > argc)
    {
        usage (argv[0]);
        result = -1;
    }

    return result;
}

/*!
 * @fn          void execute_command (int num)
 * @brief       exeucte command on button press
 * @details     exceute /usr/bin/hwsupp_cmd <num>
 *              with button number <num>
 *              the user is responsible to check the arg <num> in the
 *              hwsupp_cmd script
 */
void execute_command (int num)
{
    int result;
    char param[10];
    char* command[3] = { "/usr/bin/hwsupp_cmd", param, NULL};
    snprintf(param, sizeof(param), "%d", num + 1);

    log_printf(LOG_INFO, "button %d pressed, executing command \"%s %s\"",
            num, command[0], command[1]);

    if (g_data.pid)
    {
        return;
    }

    g_data.pid = fork();
    switch (g_data.pid)
    {
        case -1:
            log_printf(LOG_ERR, "unable to fork: %s", strerror(errno));
            g_data.pid = 0;
            break;
        case 0:
            result = execvp(command[0], command);
            if (result < 0)
            {
                log_printf(LOG_ERR, "unable to exec %s: %s", command[0],
                    strerror(errno));
            }
            exit (1);
    }
}

/*!
 * @fn          static void check_for_zombies (void)
 * @brief       check for unterminated execute processes
 */
static void check_for_zombies (void)
{
    int status, result;

    if (!g_data.pid)
        return;

    result = waitpid(-1, &status, WNOHANG);
    if (result < 0)
    {
        log_printf(LOG_ERR, "waitpid returned %s\n", strerror(errno));
    }
    else if (result > 0)
    {
        if (WIFEXITED(status))
        {
            log_printf(LOG_INFO, "%d terminated with %d\n", result,
                    WEXITSTATUS(status));
        }
        else
        {
            log_printf(LOG_INFO, "%d terminated abnormal\n", result);
        }
        g_data.pid = 0;
    }
}

/*_oOo_*/
